# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Compilation using commands depending on mimetype or file extensions"""

import glob
import mimetypes
import operator
import pathlib
import re
import shlex

from .command import Command, MissingOption


class AutoCommand(Command):
    """Compilation using rules based on mimetypes and extensions"""

    keyword = "action.autocommand"
    priority = 50

    def get_all_config(self, path):
        """Parse config and return the setup dictionary"""
        config = {}

        # Populate using global setup
        path.config.fill_blanks(
            {
                key: value
                for key, value in self.shared.setup.items()
                if (key.startswith(self.keyword) and key != self.keyword)
            }
        )

        # Analyse config
        for key, setup in path.config.items():
            if key == self.keyword or (not key.startswith(self.keyword)):
                continue

            config[key] = setup.copy()
            if setup["command"] is None:
                raise MissingOption(self.keyword, "command")

            if setup["priority"] is None:
                config[key]["priority"] = (1, key)
            else:
                config[key]["priority"] = (0, int(setup["priority"]))

            if config[key]["extensions"] is None:
                config[key]["extensions"] = ""

            config[key]["mimetypes"] = [
                re.compile(regexp)
                for regexp in shlex.split(config[key].get("mimetypes", ""))
            ]
            config[key]["extensions"] = [
                "." + ext for ext in shlex.split(config[key].get("extensions", ""))
            ]
            if (not config[key]["mimetypes"]) and (not config[key]["extensions"]):
                config[key]["extensions"] = [key[len(self.keyword) :]]

        return config

    def get_matched_config(self, path):
        """Return the configuration associated with the given file.

        May return ``None`` if no command is associated with it.
        """
        for setup in sorted(
            self.get_all_config(path).values(), key=operator.itemgetter("priority")
        ):
            if path.from_fs.suffix in setup["extensions"]:
                return setup
            for regexp in setup["mimetypes"]:
                mime = mimetypes.guess_type(path.from_fs)[0]
                if isinstance(mime, str):
                    if regexp.match(mime):
                        return setup
        return None

    def get_targets(self, path):
        targets = []
        for item in shlex.split(path.format(self.get_matched_config(path)["targets"])):
            for target in glob.glob(
                str(path.parent.from_fs.as_posix() / pathlib.Path(item))
            ):
                targets.append(pathlib.Path(target).relative_to(path.root.from_fs))
        return targets

    def command(self, path):
        return self.get_matched_config(path)["command"]

    def match(self, value):
        return self.get_matched_config(value) is not None



# Copyright Louis Paternault 2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Some features of the `tree.Tree()` class, implemented as hooks."""

import logging
import pathlib

from ..hooks import iterhook
from . import Plugin

LOGGER = logging.getLogger("evariste")


class Tree(Plugin):
    """Some features of the `tree.Tree()` class, implemented as hooks."""

    keyword = "tree"

    @staticmethod
    @iterhook("Tree.prune_before")
    def iter_ignored(tree):
        """Iterate over ignored files."""
        # pylint: disable=too-many-branches
        if tree.is_file():
            for name in ["{}.evsignore", ".{}.evsignore"]:
                if (tree.parent.from_fs / name.format(tree.basename)).exists():
                    yield tree.parent.from_source / name.format(tree.basename)
                    yield tree.from_source
        else:  # Directory
            ignorename = tree.from_fs / ".evsignore"
            if not ignorename.exists():
                return

            yield tree.from_source / ".evsignore"

            with open(ignorename, encoding="utf8") as ignorehandler:
                for line in ignorehandler.readlines():
                    line = line.strip()
                    if not line:
                        continue
                    if line.startswith("#"):
                        continue
                    if line.startswith("/"):
                        for file in tree.walk():
                            if (
                                pathlib.Path("/")
                                / file.from_fs.relative_to(tree.from_fs)
                            ).match(line):
                                yield file.from_source
                    else:
                        for file in tree.walk():
                            if file.from_source.match(line):
                                yield file.from_source

    @staticmethod
    @iterhook("Tree.prune_after")
    def iter_depends(tree):
        """Iterate over dependencies."""
        if tree.is_file():
            if tree.report is None:
                # File has not been compiled.
                yield tree.from_source
            else:
                for path in tree.report.depends:
                    yield path.resolve().relative_to(tree.root.from_fs.resolve())



# Copyright Louis Paternault 2020-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Command line client to :mod:`evs`"""

import operator

import argdispatch

from ... import plugins


def do_list(args):
    """List available plugins."""
    parser = argdispatch.ArgumentParser(
        prog="evs plugins list",
        description=("List available plugins"),
        formatter_class=argdispatch.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="Add verbosity",
    )
    parser.add_argument(
        "--dir",
        "-d",
        action="append",
        default=[],
        help="Additional plugin directories.",
    )
    arguments = parser.parse_args(args)
    if arguments.verbose:
        formatter = (
            """{plugin_type} {plugin.keyword} {plugin.__module__} "{description}" """
        )
        print("""# Type Keyword Module Description""")
    else:
        formatter = "{plugin.keyword}"

    for plugin in sorted(
        plugins.find_plugins(arguments.dir), key=operator.attrgetter("keyword")
    ):
        if plugin.plugin_type:
            plugin_type = plugin.plugin_type
        else:
            plugin_type = "none"
        if isinstance(plugin.__doc__, str):
            description = plugin.__doc__.split("\n")[0]
        else:
            description = ""
        print(
            formatter.format(
                plugin=plugin, plugin_type=plugin_type, description=description
            )
        )


def commandline_parser():
    """Return a command line parser."""

    parser = argdispatch.ArgumentParser(
        prog="evs plugins",
        description=("Some useful plugin utilities for evariste."),
        formatter_class=argdispatch.RawTextHelpFormatter,
    )

    subparsers = parser.add_subparsers(
        title="Subcommands",
    )

    subparsers.required = True
    subparsers.dest = "subcommand"
    subparsers.add_function(do_list, command="list")

    return parser


def main():
    """Main function"""
    commandline_parser().parse_args()


if __name__ == "__main__":
    main()



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Dummy, do-nothing vcs. Used for tests.

.. autoclass:: NoneVCS
   :members:
   :special-members: __contains__
"""

import pathlib

from . import VCS


class NoneVCS(VCS):
    """Dummy vcs: Does not access any file."""

    keyword = "vcs.none"

    def walk(self):
        yield from []

    def __contains__(self, path: pathlib.Path) -> bool:
        return False

    @property
    def workdir(self) -> pathlib.Path:
        return pathlib.Path.cwd()



# Copyright Louis Paternault 2017-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Cached action. Nothing has changed since last compilation."""

import contextlib
import pathlib

from ...hooks import contexthook
from . import Action, Report


class Cached(Action):
    """Cached action. Nothing has changed since last compilation."""

    keyword = "action.cached"

    def compile(self, path):
        cached = self.shared.tree[path.from_source]["changed"]
        if cached is None:
            targets = [path.from_source]
            depends = set()
        else:
            targets = [
                pathlib.Path(filename).relative_to(path.root.from_fs)
                for filename in cached["targets"]
            ]
            depends = {
                pathlib.Path(depends) for depends in cached["depends"].keys()
            } - set([path.from_fs])
        return Report(
            path,
            targets=targets,
            depends=depends,
            success=True,
            log=self.local.tree[path.from_source],
        )

    @contexthook("File.compile")
    @contextlib.contextmanager
    def cache_log(self, tree):
        """Save the compilation log in the cache."""
        yield
        self.local.tree[tree.from_source] = tree.report.log



# Copyright Louis Paternault 2017-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Command line client to :mod:`evariste`"""

import logging
import os
import sys

from ... import errors
from ...builder import Builder
from ...evs.compile.options import get_options
from ...setup import Setup
from ...utils import ChangeDir

LOGGER = logging.getLogger("evariste")

DEFAULT_SETUP = {"changed": {"time": "fs"}}


def main():
    """Main function"""
    arguments = get_options()
    try:
        # pylint: disable=no-member
        with ChangeDir(os.path.dirname(arguments.setup)):
            # Command lines arguments are available in setup
            setup = Setup.from_file(os.path.basename(arguments.setup))
            setup.fill_blanks({"arguments": vars(arguments)})
            setup.fill_blanks(DEFAULT_SETUP)
            with Builder(setup) as builder:
                builder.compile()
    except errors.EvaristeError as error:
        LOGGER.error("Error: %s.", error)
        sys.exit(1)
    except errors.EvaristeBug as bug:
        LOGGER.error("Error: %s.", bug)
        LOGGER.error(
            "You should not see this: you just discovered a bug. Please copy "
            "the full error message and report a bug."
        )
        sys.exit(2)
    except KeyboardInterrupt:
        LOGGER.error("Aborted.")
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Raw README plugin for html renderer.

.. autoclass:: HtmlReadmeRenderer
   :members:

"""

from ...jinja2.readme import Jinja2ReadmeRenderer


class HtmlReadmeRenderer(Jinja2ReadmeRenderer):
    """Html renderer for readme files, using jinja2 template engine.

    It uses the row content of the ``html`` README.
    """

    # pylint: disable=too-few-public-methods, abstract-method

    plugin_type = "renderer.html.readme"
    keyword = "renderer.html.readme.html"
    extensions = ["html", "htm"]



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the ``libdirs`` setup option"""

import os

from evariste.plugins import PluginNotFound

from . import TestLoadedPlugins


def absdir(path):
    """Return the absolute path of the argument (relative to this file)."""
    return os.path.join(os.path.dirname(__file__), "libdirs", path)


class TestLibdirs(TestLoadedPlugins):
    """Testing setup `libdirs` options"""

    # pylint: disable=too-few-public-methods

    def test_nolibdirs(self):
        """Test compilation without the libdirs option"""
        self.assertSetEqual(self._loaded_plugins({}), self.mandatory_plugins)

    def test_fail(self):
        """Test failure (plugin not found)."""
        with self.assertRaises(PluginNotFound):
            self._loaded_plugins(
                {
                    "setup": {
                        "libdirs": absdir("plugins"),
                        "plugins": ["foo", "bar"],
                    }
                }
            )

    def test_libdirs(self):
        """Test compilation with the libdirs option"""
        with self.subTest("Without spaces"):
            self.assertSetEqual(
                self._loaded_plugins(
                    {"setup": {"libdirs": absdir("plugins"), "plugins": ["foo"]}}
                ),
                set(["foo"]) | self.mandatory_plugins,
            )

        with self.subTest("With spaces"):
            self.assertSetEqual(
                self._loaded_plugins(
                    {
                        "setup": {
                            "libdirs": f'"{absdir("plug ins")}"',
                            "plugins": ["bar"],
                        }
                    }
                ),
                set(["bar"]) | self.mandatory_plugins,
            )

        with self.subTest("With and without spaces"):
            self.assertSetEqual(
                self._loaded_plugins(
                    {
                        "setup": {
                            "libdirs": "{} {}".format(
                                absdir(r"plug\ ins"), absdir("plugins")
                            ),
                            "plugins": ["foo", "bar"],
                        }
                    }
                ),
                set(["bar", "foo"]) | self.mandatory_plugins,
            )



# Copyright Louis Paternault 2017-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Cache

Implements a caching mechanism, to be able to store data between successive
runs of Évariste.
"""
import abc
import os
import pickle

from .errors import EvaristeError
from .shared import Shared, SharedDeepDict
from .utils import expand_path


class CacheError(EvaristeError):
    """Exception related to cache."""


class AbstractCache(metaclass=abc.ABCMeta):
    """Parent class of all cache classes."""

    #: Version of cache format. If the version read from cache is different
    #: from this, the cache is discarded.
    version = 4

    def __init__(self, *, setup, builder, **data):
        self.shared = Shared(setup=setup, builder=builder, **data)

    @abc.abstractmethod
    def close(self):
        """Close cache."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def clear(cls, cachedir):
        """Delete cache.

        :arg str cachedir: Cache directory.

        This is a class method, so that it can be called without the class
        being instanciated; i.e. cache can be deleted even if it is corrupted
        and cannot be loaded.
        """
        raise NotImplementedError


class NoCache(AbstractCache):
    """Fake cache: can be used if no cache is used."""

    def close(self):
        pass

    @classmethod
    def clear(cls, cachedir=None):
        pass


class Cache(AbstractCache):
    """Read cache."""

    dataname = ["tree", "plugin"]

    def __init__(self, cachedir, *, setup=None, builder=None):
        self.cachedir = expand_path(cachedir)
        os.makedirs(self.cachedir, exist_ok=True)

        data = {}

        if self._read_cache_version() == self.version:
            for base in self.dataname:
                data[base] = self._read_data(base)

        super().__init__(setup=setup, builder=builder, **data)

    def _read_cache_version(self):
        """Return cache version. Return 0 if an error occured."""
        try:
            with open(
                self._data_filename("version"), mode="r", encoding="utf8"
            ) as file:
                return int(file.read())
        except (ValueError, FileNotFoundError):
            return 0

    def close(self):
        """Close cache: write data"""
        for base in self.dataname:
            self._write_data(base, getattr(self.shared, base))
        with open(self._data_filename("version"), mode="w", encoding="utf8") as file:
            file.write(str(self.version))

    def _data_filename(self, base):
        """Return the filename corresponding to data ``base``."""
        return os.path.join(self.cachedir, f"{base}.data")

    def _read_data(self, base):
        """Read data ``base``, and return its content.

        If data could not be read, return a default value.
        """
        try:
            with open(self._data_filename(base), "rb") as file:
                return SharedDeepDict.from_attr(base, pickle.load(file))
        except (FileNotFoundError, ValueError, EOFError, pickle.PickleError):
            # File does not exist, or data is corrupted: provide default data.
            return SharedDeepDict.from_attr(base)

    def _write_data(self, base, data):
        """Write data `base`."""
        with open(self._data_filename(base), "wb") as file:
            pickle.dump(vars(data), file)

    @classmethod
    def clear(cls, cachedir):
        data = [
            os.path.join(cachedir, f"{base}.data")
            for base in ["version", "plugin", "tree"]
        ]
        try:
            for path in data:
                try:
                    os.remove(path)
                except FileNotFoundError:
                    pass
            try:
                os.rmdir(cachedir)
            except FileNotFoundError:
                pass
        except OSError as error:
            raise CacheError(f"Unable to delete '{error.filename}': {error}") from error


def open_cache(cachedir, setup, builder):
    """Return a cache object, possibly fake if `cachedir` is None."""
    if cachedir is None:
        return NoCache(setup=setup, builder=builder)
    return Cache(cachedir, setup=setup, builder=builder)



# Copyright Louis Paternault 2015-2021
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test involving compilation"""

import collections
import difflib
import filecmp
import os
import re
import sys
import unittest

from evariste.builder import Builder
from evariste.setup import Setup


class Compare:
    """Compare some characteristics of files."""

    # pylint: disable=too-few-public-methods

    message = ""

    @staticmethod
    def compare(first, second):
        """Compare files. Return `True` iff they differ."""
        raise NotImplementedError()


class FirstOnly(Compare):
    """File is present in left directory only."""

    # pylint: disable=abstract-method, too-few-public-methods

    message = "Only in first"


class SecondOnly(Compare):
    """File is present in right directory only."""

    # pylint: disable=abstract-method, too-few-public-methods

    message = "Only in second"


class SameContent(Compare):
    """Files have same content"""

    # pylint: disable=too-few-public-methods

    message = "File content differ"

    @staticmethod
    def compare(first, second):
        result = filecmp.cmp(first, second)
        if not result:
            with open(first, encoding="utf8") as leftfile:
                with open(second, encoding="utf8") as rightfile:
                    sys.stderr.writelines(
                        difflib.Differ().compare(
                            leftfile.readlines(), rightfile.readlines()
                        )
                    )
        return result


class TestCompilation(unittest.TestCase):
    """Test of compilation"""

    @staticmethod
    def _run_evariste(local):
        """Create a builder, from a dummy setup file.

        Return a set a loaded plugins.
        """
        setup = Setup({"setup": {"vcs": "vcs.git"}, "vcs": {"plugins": ["vcs.git"]}})
        setup.update(local, extend_list=True)

        with Builder.from_setupdict(setup) as builder:
            builder.compile()  # pylint: disable=no-member
            for renderer in builder.iter_renderers():  # pylint: disable=no-member
                renderer()

    def assertTreeEqual(self, first, second, rules=None):
        """Compare if directory trees `first` and `second` are equal.

        :param list rules: List of tuples `(regexp, Compare)` objects, for
            special rules concerning objects. When comparing a file, the first
            tuple with a regexp matching the file (base) name is considered,
            and its associated `Compare` class is used to compare files. By
            default, the :class:`SameContent` class is used. This is useful to
            define custom matches (for instance, two generated PDF will not be
            exactly equal, so a fuzzy comparison may be used).
        """
        # pylint: disable=invalid-name
        if rules is None:
            rules = []
        rules.append((r".*", SameContent))
        rules = [(re.compile(regexp), comparator) for regexp, comparator in rules]
        errors = self.compare_dirs(first, second, rules)
        if sum(len(files) for files in errors.values()) != 0:
            raise AssertionError(
                "\n".join(
                    [
                        # pylint: disable=consider-using-f-string
                        "{message}:\n\t{files}".format(
                            message=error.message, files="\n\t".join(errors[error])
                        )
                        for error in errors
                    ]
                )
            )

    def compare_dirs(self, first, second, rules):
        """Compare directories, and return a dictionary of differences."""
        errors = collections.defaultdict(list)

        first_set = set(os.listdir(first))
        second_set = set(os.listdir(second))
        for filename in first_set - second_set:
            errors[FirstOnly].append(filename)
        for filename in second_set - first_set:
            errors[SecondOnly].append(filename)

        for path in first_set & second_set:
            if os.path.isdir(os.path.join(first, path)):
                for (errorname, filenames) in self.compare_dirs(
                    os.path.join(first, path), os.path.join(second, path), rules
                ).items():
                    for name in filenames:
                        errors[errorname].append(os.path.join(path, name))
            else:
                for regexp, comparator in rules:
                    if regexp.match(path):
                        if not comparator.compare(
                            os.path.join(first, path), os.path.join(second, path)
                        ):
                            errors[comparator].append(path)
                        break

        return errors



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Access to git-versionned files.

.. autoclass:: Git
   :members:
   :special-members: __contains__
"""

import datetime
import logging
import os
import pathlib
import typing
from datetime import datetime

import git

from . import VCS, NoRepositoryError

LOGGER = logging.getLogger(__name__)


class Git(VCS):
    """Access git-versionned files"""

    # pylint: disable=no-member

    keyword = "vcs.git"

    def __init__(self, shared):
        super().__init__(shared)
        try:
            self.repository = git.Repo(
                self.source.as_posix(), search_parent_directories=True
            )
        except git.InvalidGitRepositoryError as error:
            raise NoRepositoryError(self.keyword, self.source) from error

        self._last_modified = self._read_modification_date()

    def _read_modification_date(self):
        """Return a dictionary of versionned files and their last modification time."""
        # Thanks to Marian https://stackoverflow.com/a/35464230
        LOGGER.info("Reading git commit dates…")
        last_modified = {}
        for blob in self.repository.tree():
            if not os.path.isfile(blob.path):
                continue
            commit = next(self.repository.iter_commits(paths=blob.path, max_count=1))
            last_modified[pathlib.Path(blob.path)] = datetime.fromtimestamp(
                commit.committed_date
            )
        # Files not committed yet
        for blob in self.repository.index.iter_blobs():
            if blob[1].path not in last_modified:
                last_modified[pathlib.Path(blob[1].path)] = super().last_modified(
                    self.workdir / blob[1].path
                )
        LOGGER.info("Done")
        return last_modified

    def walk(self) -> typing.Iterable[pathlib.Path]:
        for entry in self._last_modified:
            try:
                yield (self.workdir / entry).relative_to(self.source)
            except ValueError:
                # `self.source` is not a subpath of `path`
                continue

    def __contains__(self, path: pathlib.Path) -> bool:
        try:
            return self.from_repo(path) in self._last_modified
        except ValueError:
            # ``path`` is not a subpath of ``self.workdir``
            return False

    @property
    def workdir(self) -> pathlib.Path:
        return pathlib.Path(self.repository.working_dir)

    def last_modified(self, path: pathlib.Path) -> datetime:
        if path not in self:
            return super().last_modified(path)

        return self._last_modified[self.from_repo(path)]



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Set of plugins testing plugin dependency, and enabling/disabling plugins.

Does nothing, but printing stuff to standard output.
"""

import os
import pathlib
from datetime import datetime

from evariste import plugins


class TestPlugin(plugins.Plugin):
    """Generic test plugin"""


################################################################################
# Dummy plugins


class Foo(TestPlugin):
    keyword = "foo"


class Bar(TestPlugin):
    keyword = "bar"


################################################################################
# Dependencies : First simple plugins


class DependsFoo(TestPlugin):
    keyword = "dependsfoo"
    depends = ["foo"]


class DependsBar(TestPlugin):
    keyword = "dependsbar"
    depends = ["bar"]


class DependsBoth(TestPlugin):
    keyword = "dependsboth"
    depends = ["foo", "bar"]


################################################################################
# Dependecies : Non-existent dependency


class NonExistentDependency(TestPlugin):
    keyword = "nonexistentdependency"
    depends = ["nonexistent"]


################################################################################
# Dependecies : Recursive dependency


class DependsFoo2(TestPlugin):
    keyword = "dependsfoo2"
    depends = ["dependsfoo1"]


class DependsFoo1(TestPlugin):
    keyword = "dependsfoo1"
    depends = ["foo"]


################################################################################
# Circular dependency


class Circular1(TestPlugin):
    keyword = "circular1"
    depends = ["circular2"]


class Circular2(TestPlugin):
    keyword = "circular2"
    depends = ["circular1"]



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the plugin dependencies"""

import unittest

from evariste.builder import Builder
from evariste.plugins import MANDATORY_PLUGINS
from evariste.setup import Setup


class TestLoadedPlugins(unittest.TestCase):
    """Load plugins"""

    mandatory_plugins = MANDATORY_PLUGINS | {"vcs.none", "logging.quiet"}

    @staticmethod
    def _loaded_plugins(local):
        """Create a builder, from a dummy setup file.

        Return a set a loaded plugins.
        """
        setup = Setup(
            {
                "setup": {
                    "vcs": "vcs.none",
                    "source": ".",
                    "plugins": ["vcs.none"],
                },
                "logging": {
                    "logger": "quiet",
                },
            }
        )
        setup.update(local, extend_list=True)

        with Builder.from_setupdict(setup) as builder:
            return set(builder.plugins.items())  # pylint: disable=no-member



# Copyright 2015-2021 Louis Paternault
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests"""

import doctest
import importlib
import pkgutil

import evariste


def iter_modules(package):
    """Itérateur sur tous les sous-modules de celui donné en argument, récursivement."""
    yield package
    for __finder, name, ispkg in pkgutil.iter_modules(
        package.__path__, package.__name__ + "."
    ):
        newpackage = importlib.import_module(name)
        yield newpackage
        if ispkg:
            yield from iter_modules(newpackage)


def load_tests(__loader, tests, __ignore):
    """Load tests (doctests)."""
    # Loading doctests
    for module in iter_modules(evariste):
        tests.addTests(doctest.DocTestSuite(module))
    return tests



# Copyright Louis Paternault 2017-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Actions performed to compile files.

The result of an :class:`action <Action>`
(Was it sucessful? Which files were used? What is the log? etc.)
is stored as a :class:`report <Report>`.

If you plan to write your own action plugin, see :ref:`write_action`.

:class:`Action`
---------------

.. autoclass:: Action
   :members:

:class:`Report`
---------------

.. autoclass:: Report
   :members:
"""

# Can be removed starting with python3.11
from __future__ import annotations

import abc
import contextlib
import io
import logging
import os
import pathlib
import threading
import typing

from ... import errors, plugins
from ...hooks import contexthook

if typing.TYPE_CHECKING:
    from ...tree import Tree

LOGGER = logging.getLogger(__name__)


################################################################################
# Actions


class Action(plugins.Plugin, metaclass=abc.ABCMeta):
    """Generic action

    Subclass this to create a new action (see :ref:`write_action`).
    """

    # pylint: disable=too-many-instance-attributes, too-few-public-methods

    plugin_type = "action"

    #: A lock shared by every action.
    #: Can be used for parts of the compilation which are not thread-safe.
    lock: threading.Lock = threading.Lock()

    @abc.abstractmethod
    def compile(self, path: Tree) -> Report:
        """Compile ``path``.

        This function *must* be thread-safe.
        It can use :attr:`Action.lock` if necessary.
        """
        raise NotImplementedError()

    def match(self, value: Tree) -> bool:
        """Return ``True`` if ``value`` can be compiled by this action."""
        # pylint: disable=unused-argument
        return False


class DirectoryAction(Action):
    """Fake action on directories."""

    # pylint: disable=abstract-method, too-few-public-methods

    keyword = "action.directory"

    def compile(self, path):
        success = self._success(path)
        if success:
            message = ""
        else:
            message = "At least one file in this directory failed."
        return Report(path, success=success, log=message, targets=[])

    @staticmethod
    def _success(path):
        """Return ``True`` if compilation of all subpath succeeded."""
        for sub in path:
            if not path[sub].report.success:
                return False
        return True

    def match(self, dummy):
        return False


################################################################################
# Reports


class Report:
    """Report of an action. Mainly a namespace with very few methods."""

    def __init__(self, path, targets=None, success=False, log=None, depends=None):
        # pylint: disable=too-many-arguments

        self.depends = depends
        if self.depends is None:
            self.depends = set()

        if log is None:
            self.log = ""
        else:
            self.log = log

        self.path = path
        if targets is None:
            self.targets = []
        else:
            self.targets = targets
        self._success = success

    @property
    def full_depends(self) -> typing.Set[pathlib.Path]:
        """Set of files this action depends on, including ``self.path``."""
        return self.depends | set([self.path.from_fs])

    @property
    def success(self) -> bool:
        """Was compilation sucessful?"""
        return self._success

    @success.setter
    def success(self, value):
        """Success setter."""
        self._success = value



# Copyright Louis Paternault 2015-2021
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Compilation using `make` binary"""

import shlex
import sys

from .command import Command

if sys.version_info >= (3, 8, 0):
    join = shlex.join
else:

    def join(split_command):
        """Backport of :func:`shlex.join` command.

        May be removed when support for python3.7 is dropped.
        """
        return " ".join(shlex.quote(item) for item in split_command)


class MakeCmd(Command):
    """Compilation using `make` binary"""

    keyword = "action.make"
    priority = 50
    default_setup = {"bin": "make", "options": ""}

    def command(self, path):
        # pylint:disable=consider-using-f-string
        return "{bin} {options} {targets}".format(
            bin=self.local.setup["bin"],
            options=self.local.setup["options"],
            targets=join(
                str(path.relative_to(path.parent)) for path in self.get_targets(path)
            ),
        )



# Copyright Louis Paternault 2016-2021
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Perform operations on cache."""

import argparse
import logging
import os
import pprint
import sys
import textwrap

from ...cache import Cache
from ...errors import EvaristeError
from ...setup import Setup
from ...utils import expand_path

LOGGER = logging.getLogger(__name__)


def _cache_type(name):
    """Check the argument and return its value.

    The argument must be an existing file or directory.
    """
    if not os.path.exists(name):
        raise argparse.ArgumentTypeError(f"File '{name}' does not exist.")
    return name


def commandline_parser():
    """Return a command line parser."""
    # pylint: disable=line-too-long

    parser = argparse.ArgumentParser(
        prog="evs cache",
        description="Perform operations on cache.",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument(
        "cache",
        metavar="CACHE",
        help=textwrap.dedent(
            """
            Cache to process.

            If neither '--setup' nor '--dir' options are provided, implicit '--setup' is used if argument is a file, and implicit '--dir' is used if argument is a directory.
            """
        ),
        type=_cache_type,
    )
    group = parent.add_mutually_exclusive_group()
    group.add_argument(
        "-s",
        "--setup",
        help="Argument is considered to be a setup file; cache directory is deduced from it.",
        action="store_true",
    )
    group.add_argument(
        "-d",
        "--dir",
        help="Argument is considered to be the cache directory.",
        action="store_true",
    )

    subparsers = parser.add_subparsers(description="", dest="command")
    subparsers.required = True

    clear = subparsers.add_parser(
        "clear", description="Delete cache.", parents=[parent], help="Delete cache."
    )
    clear.set_defaults(command=do_clear)

    console = subparsers.add_parser(
        "console",
        description="Load cache and start an interactive console, with access to this cache.",
        parents=[parent],
        help="Load cache and start an interactive console.",
    )
    console.set_defaults(command=do_console)

    dump = subparsers.add_parser(
        "dump",
        description="Dump cache. For more control over cache, see the 'console' subcommand.",
        parents=[parent],
        help="Dump cache (print its reprensentation).",
    )
    dump.set_defaults(command=do_dump)

    return parser


def cache_path(options):
    """Return the path of the cache."""
    if options.setup:
        setup = Setup.from_file(options.cache)
        if setup["setup"]["cachedir"] is None:
            return os.path.join(
                os.path.dirname(options.cache),
                f".{os.path.basename(options.cache)}.cache",
            )
        return os.path.join(
            os.path.dirname(options.cache), expand_path(setup["setup"]["cachedir"])
        )
    return options.cache


def get_cache_class(options):
    """Return the :class:`Cache` class corresponding to the setup file."""
    # pylint: disable=unused-argument
    return Cache


def get_cache(options):
    """Return the :class:`Cache` object corresponding to the setup file."""
    if options.setup:
        setup = Setup.from_file(options.cache)
    else:
        setup = None

    return get_cache_class(options)(cachedir=cache_path(options), setup=setup)


def do_console(namespace):
    """Execute the `evs cache console` command."""
    # pylint: disable=line-too-long
    cache = get_cache(namespace)

    try:
        # pylint: disable=import-outside-toplevel
        import ipdb as pdb
    except ImportError:
        print(
            """You can install the `ipdb` package to get an enhanced Python debugger."""
        )
        # pylint: disable=import-outside-toplevel
        import pdb

    # pylint: disable=unused-variable
    version = cache.version
    plugin = cache.shared.plugin  # pylint: disable=no-member
    tree = cache.shared.tree  # pylint: disable=no-member

    print(
        textwrap.dedent(
            """\
        You have access to the following variables. You can modify value of variables marked as 'read-only', but their value will not be written into cache:
        - cache: the `Cache` object.
        - version: the cache version (read-only).
        - plugin: the cached plugin information.
        - tree: the cached tree information.

        By default, nothing is written into cache. Use `cache.close()` if you want to. It is possible to write malformed cache data, which cannot be read again. If so, you can use `evs cache clear` to remove cache completely.

        Type of variables `plugin` and `tree` are custom types. To get a read-only version as a dictonary, use `vars(plugin)` and `vars(tree)`.

        Use `Ctrl-D` or `exit()` to leave the debugger.
        """
        )
    )

    # pylint: disable=forgotten-debug-statement
    pdb.set_trace()


def do_dump(namespace):
    """Execute the `evs cache dump` command."""
    cache = get_cache(namespace)

    dump = {"version": cache.version}
    dump.update({base: getattr(cache.shared, base) for base in cache.dataname})

    pprint.pprint(dump)


def do_clear(namespace):
    """Execute the `evs cache clear` command."""
    get_cache_class(namespace).clear(cache_path(namespace))


def parse_options(args):
    """Parse options, and perform some post-processing."""
    argumentparser = commandline_parser()
    options = argumentparser.parse_args(args)

    # Processing --setup/--dir options
    if options.setup:
        if not os.path.isfile(options.cache):
            argumentparser.error(
                f"Argument '{options.cache}' provided with options '--setup' must be a file."
            )
    if options.dir:
        if not os.path.isdir(options.cache):
            argumentparser.error(
                f"Argument '{options.cache}' provided with options '--dir' must be a directory."
            )
    if not (options.setup or options.dir):
        options.setup = os.path.isfile(options.cache)
        options.dir = os.path.isdir(options.cache)

    return options


def main(args=None):
    """Main function: run from command line."""
    options = parse_options(args)
    try:
        options.command(options)
    except EvaristeError as error:
        LOGGER.error(str(error))
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright Louis Paternault 2021-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Plugin log progress."""

import contextlib
import logging
import sys
import threading

from rich.console import Console
from rich.logging import RichHandler
from rich.progress import Progress

from ..hooks import contexthook
from . import Plugin

LOGGER = logging.getLogger()

LOGGING_LEVELS = {
    -1: 100,  # Quiet
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
    3: 1,
}


def get_logging_level(verbose):
    """Turn command line verbosity into :mod:`logging` verbosity."""
    if verbose in LOGGING_LEVELS:
        return LOGGING_LEVELS[verbose]
    if verbose is None:
        return LOGGING_LEVELS[0]
    if verbose > max(LOGGING_LEVELS.keys()):
        return LOGGING_LEVELS[max(LOGGING_LEVELS.keys())]
    raise NotImplementedError()


class Logging(Plugin):
    """Main plugin to log stuff.

    Its main goal is to load the right "actual" plugin depending on the configuration.
    """

    keyword = "logging"

    @classmethod
    def depends_dynamic(cls, shared):
        yield f"""logging.{shared.setup["logging"].get("logger", "auto")}"""


class LoggingQuiet(Plugin):
    """Does not log anything.

    Note that things logged *before* this plugin is enabled
    are still logged using the default Python module.
    """

    keyword = "logging.quiet"


class LoggingAuto(Logging):
    """Automatic selection of logging plugin.

    If standard output is a tty, log using nice colors and progress bar.
    Otherwise, log without frills.
    """

    keyword = "logging.auto"

    @classmethod
    def depends_dynamic(cls, shared):
        if shared.setup["logging"].get("logger", "auto") == "auto":
            if sys.stdout.isatty():
                yield "logging.rich"
            else:
                yield "logging.stdlib"


class LoggingStdlib(Logging):
    """Log stuff"""

    # pylint: disable=too-few-public-methods

    keyword = "logging.stdlib"
    default_setup = {"format": "%(message)s"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lock = threading.Lock()
        self.count = 0
        self.total = 0

        LOGGER.addHandler(self.get_handler())
        self.set_level(LOGGER)

    def get_handler(self):
        """Return the :class:`logging.Handler` used by this class."""
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(self.local.setup["format"]))
        return handler

    def set_level(self, logger):
        """Process --verbose and --quiet options."""
        arguments = dict(self.shared.setup["arguments"])
        if arguments.get("quiet", None):
            arguments["verbose"] = -1
        logger.setLevel(get_logging_level(arguments.get("verbose", 0)))

    @contexthook("File.compile")
    @contextlib.contextmanager
    def compile_file(self, tree):
        """Log file compilation (including progress)."""
        with self.lock:
            if self.count == 0:
                # Save the total number of files to compile.
                self.total = self.shared.builder.tree.count()
            self.count += 1
            count = self.count
        LOGGER.info(
            f"Compiling [{count: >{len(str(self.total))}}/{self.total}] "
            f"'{tree.from_source}'…"
        )

        yield

        if tree.report.success:
            LOGGER.info(
                f"Compiling [{count: >{len(str(self.total))}}/{self.total}] "
                f"'{tree.from_source}': success."
            )
        else:
            LOGGER.info(
                f"Compiling [{count: >{len(str(self.total))}}/{self.total}] "
                f"'{tree.from_source}': failed."
            )


class LoggingRich(LoggingStdlib):
    """Log stuff with colors and progress bar."""

    keyword = "logging.rich"

    def __init__(self, *args, **kwargs):
        self.console = Console()
        super().__init__(*args, **kwargs)

    def get_handler(self):
        return RichHandler(console=self.console)

    @contexthook("Builder.compile")
    @contextlib.contextmanager
    def compile_builder(
        self, builder, *args, **kwargs
    ):  # pylint: disable=unused-argument
        """Save the total number of files to compile."""

        # pylint: disable=attribute-defined-outside-init
        with Progress(console=self.console, transient=True) as self.progress:
            self.task = self.progress.add_task("Compiling", total=0)
            yield

    @contexthook("File.compile")
    @contextlib.contextmanager
    def compile_file(self, tree):
        """Log file compilation (including progress)."""
        with super().compile_file(tree):
            yield
            self.progress.update(self.task, advance=1, total=self.total)



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""Plugin base class and plugin loader

Every plugin is a subclass of :class:`Plugin` (see :ref:`write` for more information).

The :class:`Loader` class finds and loads the plugins.

Constants
---------

.. autodata:: MANDATORY_PLUGINS

:class:`Plugin`
---------------

.. autoclass:: Plugin
   :members:

:class:`Loader`
---------------

.. autoclass:: Loader
   :members: get_plugin, iter, values, items, applyiterhook, match

Functions
---------

.. autofunction:: find_plugins

Exceptions
----------

.. autoclass:: NoMatch

"""

import collections
import contextlib
import functools
import itertools
import logging
import os
import shlex
import sys
import typing

from .. import errors, hooks, setup, utils
from ..shared import Shared

LOGGER = logging.getLogger("evariste")

#: Set of mandatory plugins: plugins that are loaded by default, and cannot be disabled.
MANDATORY_PLUGINS = {
    "action.cached",
    "action.directory",
    "action.noplugin",
    "action.raw",
    "changed",
    "logging",
    "tree",
}


class NoMatch(errors.EvaristeError):
    """No plugin found matching ``value``."""

    def __init__(self, value, available):
        super().__init__()
        self.value = value
        self.available = available

    def __str__(self):
        return f"Value '{self.value}' does not match any of {self.available}."


class SameKeyword(errors.EvaristeError):
    """Two plugins have the same keyword."""

    def __init__(self, keyword, plugin1, plugin2):
        super().__init__()
        self.keyword = keyword
        self._plugins = (plugin1, plugin2)

    def __str__(self):
        # pylint: disable=line-too-long
        return f"""Plugins '{self._plugins[0].__name__}' and '{self._plugins[1].__name__}' have the same keyword '{self.keyword}'."""


class NotAPlugin(errors.EvaristeError):
    """Superclass of plugins is not a plugin."""

    def __init__(self, obj):
        super().__init__()
        self.obj = obj

    def __str__(self):
        return (
            """Class '{obj.__module__}.{obj.__name__}' is not a plugin """
            """(it should inherit from """
            """'{superclass.__module__}.{superclass.__name__}')."""
        ).format(obj=self.obj, superclass=Plugin)


class PluginNotFound(errors.EvaristeError):
    """Plugin cannot be found."""

    def __init__(self, keyword):
        super().__init__()
        self.keyword = keyword

    def __str__(self):
        return f"Cannot find plugin '{self.keyword}'."


@functools.total_ordering
class Plugin:
    """Plugin base: all imported plugins must be subclasses of this class.

    See :ref:`write` to see how to write a new plugin.

    :param Shared shared: The object :ref:`shared <shared>` among plugins.
    """

    # pylint: disable=too-few-public-methods

    #: Keyword plugin, used to reference it: it is used
    #: to :ref:`enable plugins in the setup file <plugins>`,
    #: to name its section in the :ref:`setup file <setup>`, etc.
    keyword: typing.Union[None, str] = None

    #: When Évariste has to choose *one* plugin among several one,
    #: it chooses the one with higher priority.
    priority: int = 0

    #: Default value for section ``self.keyword`` in the :ref:`setup file <setup>`.
    #: It may be overwritten by data provided by user in the :ref:`setup file <setup>`.
    #: See :ref:`default_setup`.
    default_setup: typing.Dict[str, str] = {}

    #: Default values for setup file. See :ref:`default_setup`.
    global_default_setup: typing.Dict[str, typing.Dict[str, str]] = {}

    #: Type of the plugin.
    #: Plugins of the same type gather some common behaviour.
    plugin_type: str = ""

    #: Iterable of plugins this plugin depends on.
    #: When this plugin is enabled, those plugins are enabled as well.
    depends: typing.Iterator[str] = ()

    def __init__(self, shared):
        #: Common data shared with every :class:`Tree` and :class:`~evariste.plugins.Plugin`
        #: of this :class:`~evariste.builder.Builder`.
        self.shared: Shared = shared

        #: Same as :attr:`Plugin.shared`, but from this plugin point of view:
        #: see :meth:`evariste.shared.Shared.get_plugin_view` and :ref:`plugin_local`.
        self.local = shared.get_plugin_view(self.keyword)

        self._set_default_setup()

    @classmethod
    def depends_dynamic(cls, shared) -> typing.Iterator[str]:
        """Iterator of plugins this plugin depends on (as an iterator of :class:`str`)

        When this plugin is enabled, those plugins are enabled as well.

        :param Shared shared: Shared object of the current builder.

        Warning: called before everything is settled down
        """
        # pylint: disable=unused-argument
        yield from ()

    def _set_default_setup(self):
        """Set default value for this plugin setup, if necessary."""
        default = setup.Setup()
        for parent in reversed(self.__class__.mro()):  # pylint: disable=no-member
            if hasattr(parent, "global_default_setup"):
                default.update(parent.global_default_setup)
            if hasattr(parent, "default_setup"):
                default.update({self.keyword: parent.default_setup})
        self.shared.setup.fill_blanks(default)

    def match(self, value, *args, **kwargs) -> bool:  # pylint: disable=unused-argument
        """Return ``True`` iff ``value`` matches ``self``.

        Default is keyword match.
        This method can be overloaded by subclasses.
        """
        return value == self.keyword

    def __lt__(self, other):
        priority = self.priority
        if callable(priority):
            priority = priority()  # pylint: disable=not-callable
        other_priority = other.priority
        if callable(other_priority):
            other_priority = other_priority()
        if priority == other_priority:
            return self.keyword < other.keyword
        return priority < other_priority


def find_plugins(
    libdirs: typing.Union[typing.Iterable[str], None] = None
) -> typing.Iterator[Plugin]:
    """Iterate over available plugins.

    :param typing.Iterable[str] libdirs:
        Additional iterable of directories where plugins can be found.
    """
    if libdirs is None:
        libdirs = []

    path = []
    path.extend(
        [
            os.path.join(utils.expand_path(item), "plugins")
            for item in [".evariste", "~/.local/evariste", "~/.evariste"]
        ]
    )
    path.extend([utils.expand_path(item) for item in libdirs])
    path.extend([os.path.join(item, "evariste") for item in sys.path])

    yielded = set()
    for module in utils.iter_modules(path, "evariste."):
        for attr in dir(module):
            if attr.startswith("_"):
                continue
            obj = getattr(module, attr)

            if isinstance(obj, type) and issubclass(obj, Plugin):
                if obj.keyword is None:
                    continue
                if obj in yielded:
                    continue
                yielded.add(obj)
                yield obj


def find_plugins_sorted(libdirs=None):
    """Like :fun:`find_plugins`, but returns a :class:`utils.DeepDict`.

    The returned object is a dictionary of dictionaries:
    - first keys are plugin types;
    - second keys are plugin keywords;
    - values are the only plugin of said type and said keyword.
    """
    plugindict = utils.DeepDict(2)

    for plugin in find_plugins(libdirs):
        if plugin.keyword in plugindict[plugin.plugin_type]:
            raise SameKeyword(
                plugin.keyword,
                plugin,
                plugindict[plugin.plugin_type][plugin.keyword],
            )

        plugindict[plugin.plugin_type][plugin.keyword] = plugin
    return plugindict


def get_libdirs(libdirs):
    """Convert `libdirs` setup option (as a string) to a list of path (as strings)."""
    if not libdirs:
        return []
    if isinstance(libdirs, str):
        return shlex.split(libdirs)
    if isinstance(libdirs, list):
        return libdirs
    return []


class Loader:
    """Load plugins

    :param evariste.shared.Shared shared: The :ref:`shared <shared>` object among plugins.

    The constructor (:meth:`Loader.__init__`):

    - reads the :ref:`setup file <setup>` (looking for the :ref:`libdirs <libdirs>` option);
    - search all plugins (subclasses of :class:`Plugin`);
    - instanciate:

      - the :const:`mandatory plugins <MANDATORY_PLUGINS>`,
      - those enabled in the :ref:`setup file <setup>`,
      - and their dependencies;

    - store them in some attribute, so that they can be accessed later.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, *, shared):
        # pylint: disable=too-many-branches
        self.shared = shared

        # Load plugins:
        # Given the avaialble plugins, only loads relevant plugins.
        to_load = set()
        self.available = find_plugins_sorted(
            get_libdirs(self.shared.setup["setup"]["libdirs"])
        )

        # Step 0: Convert setup options in a list of keyword
        if self.shared.setup["setup"]["plugins"] is None:
            self.shared.setup["setup"]["plugins"] = []
        else:
            if isinstance(self.shared.setup["setup"]["plugins"], str):
                self.shared.setup["setup"]["plugins"] = shlex.split(
                    self.shared.setup["setup"]["plugins"]
                )
            elif isinstance(self.shared.setup["setup"]["plugins"], list):
                pass
            else:
                raise ValueError(
                    (
                        # pylint: disable=consider-using-f-string, line-too-long
                        "'Setup[setup][plugins]' should be a string or a list (is now {}: '{}')."
                    ).format(
                        type(self.shared.setup["setup"]["plugins"]),
                        self.shared.setup["setup"]["plugins"],
                    )
                )

        # Step 1: Add enabled plugins
        for keyword in self.shared.setup["setup"]["plugins"]:
            try:
                self.available.get_subkey(keyword)
            except KeyError as error:
                raise PluginNotFound(keyword) from error
            to_load.add(keyword)

        # Step 2: Enable/disable plugins according to their "enable" configuration
        for types in self.available.values():
            for plugin in types.values():
                if "enable" in self.shared.setup[plugin.keyword]:
                    if utils.yesno(self.shared.setup[plugin.keyword]["enable"]):
                        to_load.add(plugin.keyword)
                    elif plugin.keyword in MANDATORY_PLUGINS:
                        LOGGER.warning(
                            f"""Attemp to disable mandatory plugin "{plugin.keyword}" ignored."""
                        )
                        continue
                    else:
                        to_load.discard(plugin.keyword)

        # Step 3: Add default plugins
        to_load |= MANDATORY_PLUGINS

        # Step 4: Manage dependencies
        to_process = to_load.copy()
        processed = set()
        while to_process:
            keyword = to_process.pop()

            try:
                # Check if plugin exists
                self.available.get_subkey(keyword)
            except KeyError as error:
                raise PluginNotFound(keyword) from error

            for dependency in itertools.chain(
                self.available.get_subkey(keyword).depends,
                self.available.get_subkey(keyword).depends_dynamic(self.shared),
            ):
                try:
                    # Check if plugin exists
                    self.available.get_subkey(dependency)
                except KeyError as error:
                    raise PluginNotFound(dependency) from error
                if (dependency not in to_load) and (dependency not in processed):
                    to_load.add(dependency)
                    to_process.add(dependency)
            processed.add(keyword)

        # Step 5: Instanciate plugins
        self._plugins = collections.defaultdict(dict)
        self._hooks = collections.defaultdict(lambda: collections.defaultdict(set))
        for keyword in to_load:
            plugin = self.available.get_subkey(keyword)(self.shared)
            self._plugins[plugin.plugin_type][keyword] = plugin
            for hooktype, name, function in hooks.iter_hooks(plugin):
                self._hooks[hooktype][name].add(function)

    def values(
        self, plugin_type: typing.Optional[str] = None
    ) -> typing.Iterable[Plugin]:
        """Iterate over plugins.

        :param typing.Optional[str] plugin_type: See :meth:`Loader.iter`.
        """
        if plugin_type is None:
            for ptype in self._plugins:
                yield from self._plugins[ptype].values()
        else:
            yield from self._plugins[plugin_type].values()

    def applymethodhook(
        self, hookname: str, function: typing.Callable, *args, **kwargs
    ):
        """Apply a :ref:`method hook <methodhook>`.

        That is, run stuff before and after a method call.

        :param str hookname: Name of the hook to apply.
        :param function function: Function to which the hook is to be applied.
        :param list args: List of parameters to pass to hooks and ``function``.
        :param dict kwargs: Dictionary of parameters to pass to hooks and ``function``.
        """
        with contextlib.ExitStack() as stack:
            for hook in sorted(self._hooks["contexthook"][hookname], key=str):
                stack.enter_context(hook(*args, **kwargs))
            for hook in sorted(self._hooks["methodhook"][hookname], key=str):
                function = hook(function)

            return function(*args, **kwargs)

    def applyiterhook(self, hookname: str, *args, **kwargs) -> typing.Iterable:
        """Apply an :ref:`iteration hook <iterhook>`.

        Run every fonction hooked to this name (they should be iterators),
        and iterate over the chain of those iterators.

        :param str hookname: Name of the hook to apply.
        :param list arg: Positional arguments passed to the hooks.
        :param dict kwargs: Named arguments passed to the hooks.
        """
        for hook in sorted(self._hooks["iterhook"][hookname], key=str):
            yield from hook(*args, **kwargs)

    def get_plugin(self, keyword: str) -> Plugin:
        """Return the plugin with the given keyword.

        :raises NoMatch: If no (loaded) plugin was found with this keyword.
        """
        for plugin in self.values():
            if plugin.keyword == keyword:
                return plugin
        raise NoMatch(keyword, sorted(plugin.keyword for plugin in self.values()))

    def items(
        self, plugin_type: typing.Optional[str] = None
    ) -> typing.Iterable[typing.Tuple[str, Plugin]]:
        """Iterate over plugin keywords.

        :param typing.Optional[str] plugin_type: See :meth:`Loader.iter`.
        """
        for plugin in self.values(plugin_type):
            yield plugin.keyword

    def match(self, plugin_type: typing.Optional[str], value) -> Plugin:
        """Return the first plugin matching ``value``.

        A plugin ``Foo`` matches ``value`` if ``Foo.match(value)`` returns ``True``.

        :param typing.Optional[str] plugin_type: See :meth:`Loader.iter`.
        """
        for plugin in sorted(self.values(plugin_type), reverse=True):
            if plugin.match(value):
                return plugin
        raise NoMatch(value, sorted(self.iter(plugin_type)))

    def iter(
        self, plugin_type: typing.Union[str, None] = None
    ) -> typing.Iterable[Plugin]:
        """Iterate over keywords.

        :param typing.Optional[str] plugin_type: Type of the plugins to iterate over.

            - if ``None``, iterate over keywords of every (loaded) plugins;
            - else, iterate over keywords of plugins of this given type only.
        """
        if plugin_type is None:
            for ptype in self._plugins:
                yield from self._plugins[ptype]
        else:
            yield from self._plugins[plugin_type]



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the plugin dependencies"""

import os

from evariste.plugins import PluginNotFound

from . import TestLoadedPlugins

DEPENDSDIR = os.path.join(os.path.dirname(__file__), "depends")


def string2set(string):
    """Turn a string into a set of stripped non-empty lines"""
    return {line.strip() for line in string.split("\n") if line.strip()}


class TestDefaultSetup(TestLoadedPlugins):
    """Testing default setup"""

    # pylint: disable=too-few-public-methods

    def test_noplugins(self):
        """Test default and required plugins."""
        self.assertSetEqual(
            self._loaded_plugins({"setup": {"libdirs": DEPENDSDIR}}),
            self.mandatory_plugins,
        )

    def test_disable_default(self):
        """Disabling a default plugin is ignored"""
        self.assertSetEqual(
            self._loaded_plugins(
                {
                    "setup": {
                        "libdirs": DEPENDSDIR,
                    },
                    "action.row": {
                        "enable": False,
                    },
                }
            ),
            self.mandatory_plugins,
        )

    def test_non_existent_plugin(self):
        """Test exception when trying to enable a non-existent plugin."""
        with self.subTest():
            with self.assertRaises(PluginNotFound):
                self._loaded_plugins(
                    {
                        "setup": {
                            "libdirs": DEPENDSDIR,
                            "plugins": ["nonexistent"],
                        }
                    }
                )

        with self.subTest():
            with self.assertRaises(PluginNotFound):
                self._loaded_plugins(
                    {
                        "setup": {
                            "libdirs": DEPENDSDIR,
                            "plugins": ["nonexistentdependency"],
                        }
                    }
                )

    def test_depends_simple(self):
        """Test simple dependencies"""
        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins(
                    {
                        "setup": {
                            "libdirs": DEPENDSDIR,
                            "plugins": ["dependsfoo"],
                        }
                    }
                ),
                set(["dependsfoo", "foo"]) | self.mandatory_plugins,
            )

        with self.subTest():
            self.assertSetEqual(
                self._loaded_plugins(
                    {
                        "setup": {
                            "libdirs": DEPENDSDIR,
                        },
                        "dependsfoo": {
                            "enable": "True",
                        },
                    }
                ),
                set(["dependsfoo", "foo"]) | self.mandatory_plugins,
            )

    def test_depends_recursive(self):
        """Test recursive dependencies"""
        self.assertSetEqual(
            self._loaded_plugins(
                {
                    "setup": {
                        "libdirs": DEPENDSDIR,
                        "plugins": ["dependsfoo2"],
                    }
                }
            ),
            set(
                [
                    "dependsfoo1",
                    "dependsfoo2",
                    "foo",
                ]
            )
            | self.mandatory_plugins,
        )

    def test_depends_enable(self):
        """Test precedence of "enable=True" over "plugins"."""

        self.assertSetEqual(
            self._loaded_plugins(
                {
                    "setup": {
                        "libdirs": DEPENDSDIR,
                        "plugins": ["foo"],
                    },
                    "foo": {
                        "enable": "False",
                    },
                }
            ),
            self.mandatory_plugins,
        )

    def test_circular_dependency(self):
        """Test circular dependency."""

        self.assertSetEqual(
            self._loaded_plugins(
                {
                    "setup": {
                        "libdirs": DEPENDSDIR,
                        "plugins": ["circular1"],
                    },
                    "foo": {
                        "enable": "False",
                    },
                }
            ),
            {"circular1", "circular2"} | self.mandatory_plugins,
        )



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Default HTML file renderer

.. autoclass:: HtmlFileRenderer
   :members:

"""

from ...jinja2.file import Jinja2FileRenderer


class HtmlFileRenderer(Jinja2FileRenderer):
    """Default HTML file renderer.

    This displays the file name together with the file source.

    To write another file renderer, you can:

    - define a new
      :attr:`~evariste.plugins.renderer.jinja2.file.Jinja2FileRenderer.template` to use:
    - overwrite the default
      :meth:`~evariste.plugins.renderer.jinja2.file.Jinja2FileRenderer.match` method;
    - overwrite the default
      :meth:`~evariste.plugins.renderer.jinja2.file.Jinja2FileRenderer.render` method.
    """

    # pylint: disable=too-few-public-methods

    plugin_type = "renderer.html.file"
    keyword = "renderer.html.file.default"
    extension = "html"



# Copyright Louis Paternault 2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Some plugins to assist Évariste development."""

import contextlib

from ..hooks import contexthook, iterhook
from . import Plugin


class DebugHooks(Plugin):
    """Print hook calls"""

    # pylint: disable=missing-function-docstring

    keyword = "debug.hooks"

    def __init__(self, *args, **kwargs):
        print(f"Calling __init__({self}, {args}, {kwargs})")
        super().__init__(*args, **kwargs)

    @contexthook("Builder.compile")
    @contextlib.contextmanager
    def builder_compile(self, *args, **kwargs):
        print(f'Entering hook "Builder.compile": {self}, {args}, {kwargs}')
        yield
        print(f'Leaving hook "Builder.compile": {self}, {args}, {kwargs}')

    @contexthook("Builder.close")
    @contextlib.contextmanager
    def builder_close(self, *args, **kwargs):
        print(f'Entering hook "Builder.close": {self}, {args}, {kwargs}')
        yield
        print(f'Leaving hook "Builder.close": {self}, {args}, {kwargs}')

    @contexthook("Tree")
    @contextlib.contextmanager
    def tree(self, *args, **kwargs):
        print(f'Entering hook "Tree": {self}, {args}, {kwargs}')
        yield
        print(f'Leaving hook "Tree": {self}, {args}, {kwargs}')

    @contexthook("File.compile")
    @contextlib.contextmanager
    def file_compile(self, *args, **kwargs):
        print(f'Entering hook "File.compile": {self}, {args}, {kwargs}')
        yield
        print(f'Leaving hook "File.compile": {self}, {args}, {kwargs}')

    @contexthook("File.make_archive")
    @contextlib.contextmanager
    def file_make_archive(self, *args, **kwargs):
        print(f'Entering hook "File.make_archive": {self}, {args}, {kwargs}')
        yield
        print(f'Leaving hook "File.make_archive": {self}, {args}, {kwargs}')

    @iterhook("Tree.prune_before")
    def tree_prune_before(self, *args, **kwargs):
        print(f"""Calling hook "Tree.prune_before": {self}, {args}, {kwargs}""")
        yield from ()

    @iterhook("Tree.prune_after")
    def tree_prune_after(self, *args, **kwargs):
        print(f"""Calling hook "Tree.prune_after": {self}, {args}, {kwargs}""")
        yield from ()



# Copyright Louis Paternault 2016-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Directory representation and compilation.

A :class:`Tree` is an abstract class representing a directory structure
(a directory with files and nested directories).
Its implementations are:

    - :class:`File`: a file;
    - :class:`Directory`: a directory;
    - :class:`Root`: the root directory being processed.

:class:`Tree`
-------------

.. autoclass:: Tree
   :members:

:class:`File`
-------------

.. autoclass:: File
   :members:

:class:`Directory`
------------------

.. autoclass:: Directory
   :members:
   :special-members: __iter__, __contains__, __getitem__, __delitem__

:class:`Root`
-------------

.. autoclass:: Root
   :members:

"""

# Can be removed starting with python3.11
from __future__ import annotations

import abc
import datetime
import functools
import itertools
import logging
import os
import pathlib
import shlex
import tarfile
import traceback
import typing
from concurrent.futures import ThreadPoolExecutor

from . import errors, plugins, utils
from .hooks import setmethodhook
from .plugins.action import DirectoryAction

if typing.TYPE_CHECKING:
    from .shared import Shared

LOGGER = logging.getLogger(__name__)


def get_tree_plugins(self, *args, **kwargs):
    """Given the argument of :method:`Tree.__init__`, return a :class:`Loader` instance."""
    # pylint: disable=unused-argument
    if isinstance(self, Root):
        root = self
    else:
        root = kwargs["parent"]
        while root.parent is not None:
            root = root.parent
    return root.shared.builder.plugins


@functools.total_ordering
class Tree(metaclass=abc.ABCMeta):
    """A file system tree.

    A directory, that contains files and has subdirectories.

    :param pathlib.Path path: Relative path (relative to the root of this tree).
    :param typing.Optional[Directory] parent: Directory containing this file or directory.

    """

    # pylint: disable=too-many-instance-attributes

    @setmethodhook(getter=get_tree_plugins)
    def __init__(
        self, path: pathlib.Path, *, parent: typing.Optional[Directory] = None
    ):
        #: Once the file has been :meth:`compiled <File.compile>`,
        #: the report (compilation log, if any) is saved here.
        self.report: typing.Union[plugins.action.Report, None] = None

        #: Parent directory (copied from constructor argument).
        self.parent: typing.Union[Tree, None] = parent

        #: Computed configuration for this file. See :ref:`evsconfig`.
        #: Note: This attribute is :class:`None` until :meth:`Tree.set_config` has been called.
        self.config: typing.Union[utils.DeepDict, None] = None

        if parent is not None:
            #: Name of the tree (path, relative to its :attr:`Tree.parent`).
            self.basename: pathlib.Path = pathlib.Path(path)

            #: Absolute path
            self.from_fs: pathlib.Path = parent.from_fs / self.basename

            #: Path, relative to the :class:`Root`.
            self.from_source: pathlib.Path = parent.from_source / self.basename

            #: VCS plugin
            self.vcs: plugins.vcs.VCS = self.parent.vcs

            #: Common data shared with every :class:`Tree` and :class:`~evariste.plugins.Plugin`
            #: of this :class:`~evariste.builder.Builder`.
            self.shared: Shared = self.parent.shared

        #: Same as :attr:`Tree.shared`, but from a tree point of view:
        #: see :meth:`~evariste.shared.Shared.get_tree_view`.
        self.local = self.shared.get_tree_view(path)

    def __hash__(self):
        return hash((hash(self.parent), self.from_fs))

    @property
    def relativename(self) -> pathlib.Path:
        """Return a *relative* name.

        * For root, return path relative to file system (or directory of setup file).
        * For non-root, return path relative to parent (i.e. basename of path).
        """
        if isinstance(self, Root):
            return self.from_fs
        return self.basename

    @staticmethod
    def is_root():
        """Return ``True`` iff ``self`` is the root."""
        return False

    @property
    def depth(self) -> int:
        """Return the depth of the path.

        The root has depth 0, and depth of each path is one more than the depth
        of its parent.
        """
        if isinstance(self, Root):
            return 0
        return 1 + self.parent.depth

    @property
    def root(self) -> Root:
        """Return the root of the tree."""
        if self.is_root():
            return self
        return self.parent.root

    def find(
        self, path: typing.Union[str, pathlib.Path, typing.Tuple[str]]
    ) -> typing.Union[Tree, False]:
        """Return the tree object corresponding to ``path`` if it exists; False otherwise.

        Argument can be:

        - a string (:class:`str`);
        - a :class:`pathlib.Path` object;
        - a tuple of strings, as a list of directories and (optional) final file.
        """
        if isinstance(path, str):
            pathtuple = [path]
        elif isinstance(path, pathlib.Path):
            pathtuple = path.parts
        else:
            pathtuple = path

        if not pathtuple:
            return self
        if pathtuple[0] in self:
            return self[pathtuple[0]].find(pathtuple[1:])
        return False

    def __str__(self):
        return self.from_fs.as_posix()

    def __eq__(self, other):
        return self.from_source == other.from_source

    def __lt__(self, other):
        if isinstance(self, Directory) and not isinstance(other, Directory):
            return True
        if not isinstance(self, Directory) and isinstance(other, Directory):
            return False
        return self.from_source < other.from_source

    def is_dir(self) -> bool:
        """Return `True` iff `self` is a directory."""
        return issubclass(self.__class__, Directory)

    def is_file(self) -> bool:
        """Return `True` iff `self` is a file."""
        return issubclass(self.__class__, File)

    def walk(self, dirs: bool = False, files: bool = True) -> typing.Iterable[Tree]:
        """Iterator over itself."""
        #  pylint: disable=unused-argument
        if files and self.is_file():
            yield self

    def count(self, dirs: bool = False, files: bool = True) -> int:
        """Count the number of files or directories in this tree."""
        return sum(1 for _ in self.walk(dirs, files))

    def prune(self, path: typing.Union[pathlib.Path, str, typing.Tuple[str]]):
        """Remove a file.

        Argument can be either:

        - a :class:`pathlib.Path`,
        - a :class:`tuple`,
        - or a :class:`str` (which would be converted to a :class:`pathlib.Path`.

        If called with a non-existing path, does nothing.
        """
        if isinstance(path, str):
            path = pathlib.Path(path)
        if isinstance(path, tuple):
            parts = path
        elif isinstance(path, pathlib.Path):
            parts = path.parts
        else:
            raise TypeError

        # Does path exist?
        if parts[0] not in self:
            return

        # Remove (existing) path
        if len(parts) == 1:
            del self[parts[0]]
        else:
            self[parts[0]].prune(parts[1:])

    def format(self, string: str) -> str:
        """Format given string, with several variables related to ``self``.

        Here are the replacements (with example :file:`/home/louis/repo/foo/bar.txt`):

        - ``{dirname}`` (:file:`/home/louis/repo/foo`): the name of the directory.
          Note that most of the time, this is useless, since when compiling a file,
          the working directory is the directory of this file (i.e. ``{dirname}``).
        - ``{filename}`` (:file:`bar.txt`): The file name (without directory).
        - ``{fullname}`` (:file:`/home/louis/repo/foo/bar.txt`): The file name (with directory).
        - ``{extension}`` (:file:`txt`): The extension (without the dot).
          If the file has several extensions (e.g. :file:`foo.tar.gz`),
          this is only the last one ``gz``.
        - ``{basename}`` (:file:`bar`): The file name, without directory and extension.
        """
        suffix = self.from_source.suffix
        if suffix.startswith("."):
            suffix = suffix[1:]
        return string.format(
            dirname=self.parent.from_fs.as_posix(),
            filename=self.basename,
            fullname=self.from_fs.as_posix(),
            extension=suffix,
            basename=self.basename.stem,
        )

    @abc.abstractmethod
    def full_depends(self) -> typing.Iterable[pathlib.Path]:
        """Iterate over all dependencies of this tree (recursively for directories)."""
        raise NotImplementedError()


class Directory(Tree):
    """Directory"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._subpath = {}

    def add_subpath(self, sub: typing.List[pathlib.Path]):
        """Add a path to the tree (relative to ``self``)."""
        if len(sub.parts) > 1:
            self[sub.parts[0]].add_subpath(pathlib.Path(*sub.parts[1:]))
        else:
            self[sub.parts[0]]  # pylint: disable=pointless-statement

    def __iter__(self) -> typing.Iterable[str]:
        """Iterate over subpaths (this function is not recursive)."""
        return iter(self._subpath)

    def keys(self) -> typing.Iterable[str]:
        """Iterator over subpaths (as :class:`str` objects)."""
        yield from self._subpath.keys()

    def values(self) -> typing.Iterable[Tree]:
        """Iterator over subpaths (as :class:`Tree` objects)."""
        yield from self._subpath.values()

    def __contains__(self, key: str) -> bool:
        """Return ``True`` if key (a single file name or directory) is in this directory."""
        return key in self._subpath

    def __getitem__(self, key: str) -> Tree:
        """Return subfile or subdirectory ``self.from_fs / key``.

        If it does not exist, it is created first.
        """
        if key not in self:
            if (self.from_fs / key).is_dir():
                path_type = Directory
            else:
                path_type = File
            self._subpath[key] = path_type(key, parent=self)
        return self._subpath[key]

    def __delitem__(self, item: str):
        """Remove a subfile or subdirectory.

        If, after deletion, ``self`` is an empty directory (and is not root),
        ``self`` is remove from its parent.
        """
        pathitem = pathlib.Path(item)
        if len(pathitem.parts) != 1:
            raise errors.EvaristeBug(
                "Argument '{}' should be a single directory or file, "
                "not a 'directory/directory' or directory/file'."
            )
        if str(pathitem) in self:
            del self._subpath[str(pathitem)]
            if (not self._subpath) and self.is_dir() and self.parent is not None:
                del self.parent[self.basename]

    def walk(self, dirs: bool = False, files: bool = True) -> typing.Iterable[Tree]:
        """Iterator over files or directories  of ``self``.

        :param bool dirs: If `False`, do not yield directories.
        :param bool files: If `False`, do not yield files.

        Directories are yielded *before* subfiles and subdirectories.
        """
        if dirs and self.is_dir():
            yield self
        for sub in sorted(self):
            yield from self[sub].walk(dirs, files)

    def compile(self):
        """Compile directory."""
        self.report = DirectoryAction(self.shared).compile(self)

    def full_depends(self) -> typing.Iterable[pathlib.Path]:
        for sub in self:
            yield from self[sub].full_depends()


class File(Tree):
    """A file"""

    @setmethodhook()
    def compile(self):
        """Compile file."""
        # pylint: disable=too-many-branches

        # Find a plugin to compile the file
        # pylint: disable=unsubscriptable-object
        try:
            if not self.shared.builder.plugins.get_plugin("changed").compile(self):
                # Nothing has changed: it is useless to compile this again.
                compiler = self.shared.builder.plugins.get_plugin("action.cached")
            elif self.config["action"]["plugin"] is not None:
                compiler = self.shared.builder.plugins.get_plugin(
                    f"""action.{self.config["action"]["plugin"]}"""
                )
            else:
                compiler = self.shared.builder.plugins.match("action", self)
        except plugins.NoMatch:
            compiler = self.shared.builder.plugins.get_plugin("action.noplugin")

        # Actual compilation
        try:
            self.report = compiler.compile(self)
        except Exception as error:  # pylint: disable=broad-except
            if isinstance(error, errors.EvaristeError):
                message = str(error)
            else:
                message = "Error: Evariste internal error\n\n" + traceback.format_exc()
            LOGGER.error(message)
            self.report = plugins.action.Report(
                self,
                success=False,
                log=message,
            )

        # Add (optional) dependencies to the file
        # pylint: disable=unsubscriptable-object
        for regexp in itertools.chain(
            shlex.split(self.config[compiler.keyword].get("depends", "")),
            shlex.split(self.config["action"].get("depends", "")),
        ):
            for name in self.parent.from_fs.glob(regexp):
                if name.resolve() != self.from_fs.resolve():
                    if name in self.vcs:
                        self.report.depends.add(name)

    @setmethodhook()
    def make_archive(self, destdir: pathlib.Path) -> pathlib.Path:
        """Make an archive of ``self`` and its dependency.

        Steps are:

        - build the archive;
        - copy it to ``destdir``;
        - return the path of the archive, relative to ``destdir``.

        If ``self`` has no dependencies, consider the file as an archive.

        It can be called several times: the archive will be built only once.
        """

        def common_root(files):
            """Look for the common root of files given in argument.

            Returns a tuple of `(root, relative_files)`, where
            `relative_files` is a list of `files`, relative to the root.
            """
            files = [
                file.resolve()
                for file in files
                if file.resolve()
                .as_posix()
                .startswith(self.root.from_fs.resolve().as_posix())
            ]
            root = pathlib.Path()
            while True:
                prefixes = [path.parts[0] for path in files]
                if len(set(prefixes)) != 1:
                    break
                prefix = prefixes[0]
                files = [file.relative_to(prefix) for file in files]
                root /= prefix
            return root.relative_to(self.root.from_fs.resolve()), files

        if (
            # Compilation was successful
            self.report.success
            # File has not been compiled again
            and not self.shared.builder.plugins.get_plugin("changed").compile(self)
            # Archive has already been generated
            and "archivepath" in self.shared.tree[self.from_source]["changed"]
        ):
            return self.shared.tree[self.from_source]["changed"]["archivepath"]

        if len(self.report.full_depends) == 1:
            utils.copy(self.from_fs.as_posix(), (destdir / self.from_source).as_posix())
            return self.from_source
        archivepath = self.from_source.with_suffix(f"{self.from_source.suffix}.tar.gz")
        os.makedirs(os.path.dirname((destdir / archivepath).as_posix()), exist_ok=True)

        archive_root, full_depends = common_root(self.report.full_depends)
        with tarfile.open((destdir / archivepath).as_posix(), mode="w:gz") as archive:
            for file in full_depends:
                archive.add(
                    (self.root.from_fs / archive_root / file).as_posix(),
                    file.as_posix(),
                )
        return archivepath

    def last_modified(self) -> datetime.datetime:
        """Return the last modified date and time of ``self``."""
        return self.vcs.last_modified(self.from_fs)

    def full_depends(self) -> typing.Iterable[pathlib.Path]:
        yield from self.report.full_depends

    def depends(self) -> typing.Iterable[pathlib.Path]:
        """Iterator over dependencies of this file (but not the file itself)."""
        yield from self.report.depends


class Root(Directory):
    """Root object (directory with no parents)."""

    def __init__(self, path, *, vcs=None, shared=None):
        self.vcs = vcs
        self.shared = shared
        self.from_fs = pathlib.Path.cwd() / path
        self.from_source = pathlib.Path(".")
        self.basename = pathlib.Path(".")
        super().__init__(path)

    @staticmethod
    def _config2file(config):
        """Given a config (with a name ending in '.evsconfig', return the corresponding tree.

        If no such file is found, raise ValueError.
        """
        if str(config.basename) == ".evsconfig":
            # Directory
            return config.parent

        # File
        # First try: `config` is SOMETHING.evsconfig and SOMETHING is in the tree
        if tree := config.parent.find(str(config.basename)[: -len(".evsignore")]):
            return tree
        # Second try: `config` is .SOMETHING.evsconfig and SOMETHING is in the tree
        if tree := config.parent.find(str(config.basename)[1 : -len(".evsignore")]):
            return tree

        raise ValueError()

    def _find_config(self):
        """Iterate over configuration files for directories"""
        for config in self.walk():
            if str(config).endswith("evsconfig"):
                content = utils.read_config(config.from_fs, allow_no_value=True)
                source = content.get("setup", "source", fallback=None)
                if source is None:
                    try:
                        tree = self._config2file(config)
                    except ValueError:
                        logging.warning(
                            f"Found a configuration file '{config}' with no related file. Processing it as a normal file."  # pylint: disable=line-too-long
                        )
                        continue
                else:
                    tree = self.find((config.parent.from_source / source))
                    if not tree:
                        logging.warning(
                            f"Configuration file '{config}' refer to a non-existent source '{source}'."  # pylint: disable=line-too-long
                        )
                        continue
                yield tree, config, content

    def set_config(self):
        """Compute the configuration of each file of the tree.

        That is:

        - look for the file that configure it
          (typically ``foo.evsconfig`` is the configuration for file ``foo``),
        - load it,
        - and complete it using the recursive configuration of parent directories.
        """
        recursive = {}

        # Assign configuration to files
        for tree, configname, configcontent in list(self._find_config()):
            if utils.yesno(configcontent.get("setup", "recursive", fallback=False)):
                recursive[tree] = configcontent
                self.prune(configname.from_source)
                continue

            tree = self.find(tree.from_source)
            if tree.config is not None:
                LOGGER.warning(  # pylint: disable=logging-fstring-interpolation
                    f"""Configuration file "{configname}" has been ignored for "{tree.from_fs}": another configuration file already applies."""  # pylint: disable=line-too-long
                )
                continue

            self.prune(configname.from_source)
            tree.config = utils.DeepDict.from_configparser(configcontent)

        # Propagate recursive configuration from directories
        for directory, config in sorted(recursive.items(), reverse=True):
            for tree in self.find(directory.from_source).walk(dirs=True, files=True):
                if tree.config is None:
                    tree.config = utils.DeepDict.from_configparser(config)
                else:
                    tree.config.fill_blanks(config)

        # Set a default (empty) configuration to trees without any.
        for tree in self.walk(dirs=True, files=True):
            if tree.config is None:
                tree.config = utils.DeepDict(2)

    @staticmethod
    def is_root() -> bool:
        return True

    def root_compile(self):
        """Recursively compile files.."""
        # Set and remove configuration files
        self.set_config()

        # Prune files that are no longer needed.
        for file in list(
            itertools.chain.from_iterable(
                self.shared.builder.plugins.applyiterhook("Tree.prune_before", tree)
                for tree in self.walk(dirs=True, files=True)
            )
        ):
            self.prune(file)

        # Compile files
        with ThreadPoolExecutor(
            max_workers=self.shared.setup["arguments"]["jobs"]
        ) as pool:
            for path in self.walk(dirs=False, files=True):
                pool.submit(path.compile)

        # Prune files that are no longer needed.
        for file in list(
            itertools.chain.from_iterable(
                self.shared.builder.plugins.applyiterhook("Tree.prune_after", tree)
                for tree in self.walk(dirs=True, files=True)
            )
        ):
            self.prune(file)

        # "Compile" directories
        for path in reversed(list(self.walk(dirs=True, files=False))):
            path.compile()

    @classmethod
    def from_vcs(cls, repository: plugins.vcs.VCS) -> Root:
        """Return a directory, fully set."""
        tree = cls(repository.source, vcs=repository, shared=repository.shared)
        for path in repository.walk():
            tree.add_subpath(path)
        return tree



# Copyright Louis Paternault 2015-2021
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test of the shared plugin"""

import unittest

from evariste.builder import Builder


def default_shared():
    """Return a default, as empty as possible, :class:`evariste.shared.Shared` object."""
    return Builder.from_setupdict(
        {"setup": {"vcs": "vcs.none", "plugins": ["vcs.none"]}}
    ).shared


class TestShared(unittest.TestCase):
    """Testing Shared"""

    # pylint: disable=too-few-public-methods, no-member

    def test_setup(self):
        """Test Shared.setup"""
        shared = default_shared()

        self.assertEqual(shared.setup["plugin"]["foo"], None)
        shared.setup["plugin"]["foo"] = "bar"
        self.assertEqual(shared.setup["plugin"]["foo"], "bar")

        view = shared.get_plugin_view("plugin")
        self.assertEqual(view.setup["foo"], "bar")
        self.assertEqual(view.setup["foo1"], None)
        view.setup["foo1"] = "bar1"
        self.assertEqual(view.setup["foo1"], "bar1")
        self.assertEqual(shared.setup["plugin"]["foo1"], "bar1")

    def test_plugin(self):
        """Test Shared.plugin"""
        shared = default_shared()

        self.assertEqual(shared.plugin["plugin1"], None)
        shared.plugin["plugin1"] = "bar1"
        self.assertEqual(shared.plugin["plugin1"], "bar1")

        view1 = shared.get_plugin_view("plugin1")
        view2 = shared.get_plugin_view("plugin2")
        self.assertEqual(view1.plugin, "bar1")
        self.assertEqual(view2.plugin, None)
        view2.plugin = "bar2"
        self.assertEqual(view2.plugin, "bar2")
        self.assertEqual(shared.plugin["plugin2"], "bar2")

    def test_tree(self):
        """Test Shared.tree"""
        shared = default_shared()

        self.assertEqual(shared.tree["path1"]["plugin1"], None)
        shared.tree["path1"]["plugin1"] = "bar1"
        self.assertEqual(shared.tree["path1"]["plugin1"], "bar1")

        view = shared.get_tree_view("path1")
        self.assertEqual(view.tree["plugin1"], "bar1")
        self.assertEqual(view.tree["plugin2"], None)
        view.tree["plugin2"] = "bar2"
        self.assertEqual(view.tree["plugin2"], "bar2")
        self.assertEqual(shared.tree["path1"]["plugin2"], "bar2")

        plugin_view = shared.get_plugin_view("plugin1")
        self.assertEqual(plugin_view.tree["path1"], "bar1")

        tree2_view = shared.get_tree_view("path2")
        self.assertEqual(tree2_view.tree["plugin1"], None)
        self.assertEqual(plugin_view.tree["path2"], None)
        tree2_view.tree["plugin1"] = "bar"
        self.assertEqual(plugin_view.tree["path2"], "bar")
        self.assertEqual(tree2_view.tree["plugin1"], "bar")
        self.assertEqual(shared.tree["path2"]["plugin1"], "bar")



# Copyright Louis Paternault 2015-2022
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Build process: gather files, and compile them.

.. autoclass:: Builder
   :members:
"""

# Can be removed starting with python3.11
from __future__ import annotations

import contextlib
import logging
import typing

from . import errors, plugins
from .cache import open_cache
from .hooks import setmethodhook
from .plugins import Loader
from .setup import Setup, SetupError
from .tree import Root

LOGGER = logging.getLogger("evariste")


class Builder(contextlib.AbstractContextManager):
    """Takes care of build process. Can be used as a context.

    .. attribute:: shared
       :type: evariste.shared.Shared

       Object that is shared and accessible by every
       :class:`evariste.plugins.Plugin` and :class:`evariste.tree.Tree`.
       See :ref:`shared`.

    .. attribute:: cache
       :type: evariste.cache.Cache

       Data that is cached between compilations.
       Plugin developpers won't manipulate this attribute directly (see :ref:`shared`).

    .. attribute:: plugins
       :type: evariste.plugins.Loader

       Plugin loader: loaded plugins are gathered there.
    """

    # pylint: disable=no-member

    def __init__(self, setup):
        self.cache = open_cache(setup["setup"]["cachedir"], setup, self)
        self.shared = self.cache.shared
        self.plugins = Loader(shared=self.shared)

        LOGGER.info("Building directory tree…")
        try:
            vcs = list(self.plugins.values(plugin_type="vcs"))
            if len(vcs) != 1:
                raise errors.EvaristeError(
                    # pylint: disable=consider-using-f-string
                    "Exactly one vcs plugin must be enabled (right now: {}).".format(
                        ", ".join([plugin.keyword for plugin in vcs])
                    )
                )
            self.tree = Root.from_vcs(vcs[0])
        except plugins.NoMatch as error:
            raise SetupError(
                # pylint: disable=consider-using-f-string
                "Setup error: Value '{}' is not a valid vcs (available ones are: {}).".format(
                    error.value,
                    ", ".join([f"'{item}'" for item in error.available]),
                )
            ) from error

    @setmethodhook()
    def compile(self):
        """Compile files handled by this builder."""
        LOGGER.info("Compiling…")
        self.tree.root_compile()

    @setmethodhook()
    def close(self):
        """Perform close operations.

        Mainly used as a :ref:`methodhook`.
        """
        self.cache.close()

    @classmethod
    def from_setupname(cls, name: str) -> Builder:
        """Factory that returns a builder, given the name of a :ref:`setup file <setup>`."""
        LOGGER.info("Reading setup…")
        return cls(Setup.from_file(name))

    @classmethod
    def from_setupdict(
        cls, dictionary: typing.Dict[str, typing.Dict[str, str]]
    ) -> Builder:
        """Factory that returns a builder, given a setup dictionary.

        A *setup dictionary* is a dict that mimics :mod:`configparser` structure.
        """
        LOGGER.info("Reading setup…")
        return cls(Setup(dictionary))

    def __enter__(self, *args, **kwargs):
        # pylint: disable=useless-super-delegation
        return super().__enter__(*args, **kwargs)

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.close()
        return super().__exit__(exc_type, exc_value, traceback)
